/**********************************************************
 * view.c - view transformation module
 *
 * 1994 by Patrick J. Angeles
 *
 *
 **********************************************************/

#include <stdio.h>
#include <math.h>

#include "matrix.h"
#include "view.h"
#include "main.h"

#define RAD(x) (M_PI/180.0* (x))

/********** DEBUGing flags...
#define DEBUG_CLIP
#define DEBUG_VIEW
#define DEBUG_PIPE
*/

/********** THE FOLLOWING FOR INTERNAL USE ONLY ***********/
extern short    point_clip();
extern short    vector_clip();

static point    view_ref = {0.0, 10.0, 50.0};
static vector   right_v = {1.0, 0.0, 0.0, 1.0};
static vector   up_v = {0.0, 1.0, 0.0, 1.0};
static vector   normal_v = {0.0, 0.0, 1.0, 1.0};
static matrix   vo_matrix, pp_matrix;
       matrix   vp_matrix, T_matrix;
       double   Zmin = 0.0;
static double   Zdepth = 0.0;


void            print_vrp()
{
   printf("vrp:\n");
   print_point(&view_ref);
   printf("x,y,z:\n");
   print_vector(right_v);
   print_vector(up_v);
   print_vector(normal_v);
}

/*****************************************************
 * void init_orientation( vrp,vpn,vup)
 *   vrp - view reference point
 *   vpn - view plane normal
 *   vup - vector up
 * Initializes viewer's position and look direction.
 ****************************************************/
void            init_orientation(vrp, vpn, vup)
point          *vrp;
vector          vpn, vup;
{
   matrix          T, R;

   point_copy(&view_ref, vrp);
   unit_vector(normal_v, vpn);
   unit_vector(up_v, vup);
   cross_product(right_v, normal_v, up_v);

#ifndef TRANSFORM
   vo_matrix[0][0] = -right_v[0];
   vo_matrix[0][1] = -right_v[1];
   vo_matrix[0][2] = -right_v[2];
   vo_matrix[0][3] = 0.0;

   vo_matrix[1][0] = up_v[0];
   vo_matrix[1][1] = up_v[1];
   vo_matrix[1][2] = up_v[2];
   vo_matrix[1][3] = 0.0;

   vo_matrix[2][0] = normal_v[0];
   vo_matrix[2][1] = normal_v[1];
   vo_matrix[2][2] = normal_v[2];
   vo_matrix[2][3] = 0.0;

   vo_matrix[3][0] = -right_v[0] * -view_ref.x + up_v[0] * -view_ref.y
                       + normal_v[0] * -view_ref.z;
   vo_matrix[3][1] = -right_v[1] * -view_ref.x + up_v[1] * -view_ref.y
                       + normal_v[1] * -view_ref.z;
   vo_matrix[3][2] = -right_v[2] * -view_ref.x + up_v[2] * -view_ref.y
                       + normal_v[2] * -view_ref.z;
   vo_matrix[3][3] = 1.0;
#else
   vo_matrix[0][0] = -right_v[0];
   vo_matrix[0][1] = up_v[0];
   vo_matrix[0][2] = normal_v[0];
   vo_matrix[0][3] = -right_v[0] * -view_ref.x + up_v[0] * -view_ref.y
                       + normal_v[0] * -view_ref.z;

   vo_matrix[1][0] = -right_v[1];
   vo_matrix[1][1] = up_v[1];
   vo_matrix[1][2] = normal_v[1];
   vo_matrix[1][3] = -right_v[1] * -view_ref.x + up_v[1] * -view_ref.y
                       + normal_v[1] * -view_ref.z;

   vo_matrix[2][0] = -right_v[2];
   vo_matrix[2][1] = up_v[2];
   vo_matrix[2][2] = normal_v[2];
   vo_matrix[2][3] = -right_v[2] * -view_ref.x + up_v[2] * -view_ref.y
                       + normal_v[2] * -view_ref.z;

   vo_matrix[3][0] = 0.0;
   vo_matrix[3][1] = 0.0;
   vo_matrix[3][2] = 0.0;
   vo_matrix[3][3] = 1.0;
#endif
}

/*****************************************************
 * void eval_orientation( theta,forw,elev)
 *   theta - pan camera across
 *   forw - distance forward
 *   elev - distance up
 * Re-orients the viewer w/ respect to previous
 * orientation, and moves the viewer forward
 * along the normal_v direction (vo_matrix).
 ****************************************************/
void            eval_orientation(double theta, double elev, double forw)
{
   matrix          Ri;
   vector          temp;
   double          Angle;
   double          sin_Angle;
   double          cos_Angle;

#ifdef SUN
   Angle = theta / 180.0;
   sincospi(Angle, &sin_Angle, &cos_Angle);
#else
   sincosd(Angle, &sin_Angle, &cos_Angle);
#endif

   double2vector(normal_v, sin_Angle, 0.0, cos_Angle);
   double2vector(right_v, normal_v[2], 0.0, -normal_v[0]);
   double2vector(up_v, 0.0, 1.0, 0.0);


#ifndef TRANSFORM
   Ri[0][0] = -right_v[0];
   Ri[0][1] = up_v[0];
   Ri[0][2] = normal_v[0];
   Ri[0][3] = 0.0;

   Ri[1][0] = -right_v[1];
   Ri[1][1] = up_v[1];
   Ri[1][2] = normal_v[1];
   Ri[1][3] = 0.0;

   Ri[2][0] = -right_v[2];
   Ri[2][1] = up_v[2];
   Ri[2][2] = normal_v[2];
   Ri[2][3] = 0.0;

   Ri[3][0] = 0.0;
   Ri[3][1] = 0.0;
   Ri[3][2] = 0.0;
   Ri[3][3] = 1.0;

   double2vector(temp, 0.0, elev, forw);
   vector_transform1(temp, temp, Ri);
   point_move(&view_ref, temp);

   vo_matrix[0][0] = -right_v[0];
   vo_matrix[0][1] = -right_v[1];
   vo_matrix[0][2] = -right_v[2];
   vo_matrix[0][3] = 0.0;

   vo_matrix[1][0] = up_v[0];
   vo_matrix[1][1] = up_v[1];
   vo_matrix[1][2] = up_v[2];
   vo_matrix[1][3] = 0.0;

   vo_matrix[2][0] = normal_v[0];
   vo_matrix[2][1] = normal_v[1];
   vo_matrix[2][2] = normal_v[2];
   vo_matrix[2][3] = 0.0;

   vo_matrix[3][0] = -right_v[0] * -view_ref.x + up_v[0] * -view_ref.y
                       + normal_v[0] * -view_ref.z;
   vo_matrix[3][1] = -right_v[1] * -view_ref.x + up_v[1] * -view_ref.y
                       + normal_v[1] * -view_ref.z;
   vo_matrix[3][2] = -right_v[2] * -view_ref.x + up_v[2] * -view_ref.y
                       + normal_v[2] * -view_ref.z;
   vo_matrix[3][3] = 1.0;
#else
#if 0
   Ri[0][0] = -right_v[0];
   Ri[0][1] = -right_v[1];
   Ri[0][2] = -right_v[2];
   Ri[0][3] = 0.0;

   Ri[1][0] = up_v[0];
   Ri[1][1] = up_v[1];
   Ri[1][2] = up_v[2];
   Ri[1][3] = 0.0;

   Ri[2][0] = normal_v[0];
   Ri[2][1] = normal_v[1];
   Ri[2][2] = normal_v[2];
   Ri[2][3] = 0.0;

   Ri[3][0] = 0.0;
   Ri[3][1] = 0.0;
   Ri[3][2] = 0.0;
   Ri[3][3] = 1.0;

   double2vector(temp, 0.0, elev, forw);
   vector_transform1(temp, Ri);
   point_move(&view_ref, temp);
#else

   view_ref.x += -right_v[1] * elev + -right_v[2] * forw;
   view_ref.y += up_v[1] * elev + up_v[2] * forw;
   view_ref.z += normal_v[1] * elev + normal_v[2] * forw;

#endif

   vo_matrix[0][0] = -right_v[0];
   vo_matrix[0][1] = up_v[0];
   vo_matrix[0][2] = normal_v[0];
   vo_matrix[0][3] = -right_v[0] * -view_ref.x + up_v[0] * -view_ref.y
                       + normal_v[0] * -view_ref.z;

   vo_matrix[1][0] = -right_v[1];
   vo_matrix[1][1] = up_v[1];
   vo_matrix[1][2] = normal_v[1];
   vo_matrix[1][3] = -right_v[1] * -view_ref.x + up_v[1] * -view_ref.y
                       + normal_v[1] * -view_ref.z;

   vo_matrix[2][0] = -right_v[2];
   vo_matrix[2][1] = up_v[2];
   vo_matrix[2][2] = normal_v[2];
   vo_matrix[2][3] = -right_v[2] * -view_ref.x + up_v[2] * -view_ref.y
                       + normal_v[2] * -view_ref.z;

   vo_matrix[3][0] = 0.0;
   vo_matrix[3][1] = 0.0;
   vo_matrix[3][2] = 0.0;
   vo_matrix[3][3] = 1.0;

   mult_matrix(pp_matrix, vo_matrix, T_matrix);
#endif

#ifdef DEBUG_VIEW
   printf("view orientation matrix:\n");
   print_matrix(vo_matrix);
#endif

}

/*****************************************************
 * void eval_projection( prp,F,B,frustrum)
 *   D - depth of projection: prp = <0.0 0.0 D>.
 *   F - front clipping plane.
 *   B - back clipping plane.
 *   frustrum - in degrees, scope of vision
 * Evaluates the perspective projection matrix (pp_matrix).
 * Simplified version of Foley, VanDam projection
 *  matrix. Check it for details!
 ****************************************************/
void            eval_projection(D, F, B, frustrum)
double          D;
double          F, B;
double          frustrum;
{
   double          umin, umax, vmin, vmax;
   vector          temp;
   matrix          M, N;

   /*******************************************
    * Compute umax, etc. in terms of frustrum.
    *******************************************/
#ifdef SUN
   umax = vmax = D * tanpi(frustrum / 360.0);
#else
   umax = vmax = D * tand(frustrum / 2.0);
#endif
   umin = vmin = -umax;
#ifndef TRANSFORM
   pp_matrix[0][0] = (2.0 * -D) / ((umax - umin) * (B - D));
   pp_matrix[0][1] = 0.0;
   pp_matrix[0][2] = 0.0;
   pp_matrix[0][3] = 0.0;

   pp_matrix[1][0] = 0.0;
   pp_matrix[1][1] = (2.0 * -D) / ((vmax - vmin) * (B - D));
   pp_matrix[1][2] = 0.0;
   pp_matrix[1][3] = 0.0;

   pp_matrix[2][0] = 0.0;
   pp_matrix[2][1] = 0.0;
   pp_matrix[2][2] = -1.0 / (B - D);
   pp_matrix[2][3] = 0.0;

   pp_matrix[3][0] = 0.0;
   pp_matrix[3][1] = 0.0;
   pp_matrix[3][2] = D / (B - D);
   pp_matrix[3][3] = 1.0;
#else

   pp_matrix[0][0] = (2.0 * -D) / ((umax - umin) * (B - D));
   pp_matrix[0][1] = 0.0;
   pp_matrix[0][2] = 0.0;
   pp_matrix[0][3] = 0.0;

   pp_matrix[1][0] = 0.0;
   pp_matrix[1][1] = (2.0 * -D) / ((vmax - vmin) * (B - D));
   pp_matrix[1][2] = 0.0;
   pp_matrix[1][3] = 0.0;

   pp_matrix[2][0] = 0.0;
   pp_matrix[2][1] = 0.0;
   pp_matrix[2][2] = -1.0 / (B - D);
   pp_matrix[2][3] = D / (B - D);

   pp_matrix[3][0] = 0.0;
   pp_matrix[3][1] = 0.0;
   pp_matrix[3][2] = 0.0;
   pp_matrix[3][3] = 1.0;

#endif
   Zmin = -(F - D) / (B - D);
   Zdepth = -D;

#ifdef DEBUG_VIEW
   printf("perspective projection matrix:\n");
   print_matrix(pp_matrix);
#endif
}

/************************************************************
 * void eval_viewport( vp_minx,vp_maxx,vp_miny,vp_maxy,
 *                     vp_minz,vp_maxz)
 *   - min and max parameters in screen coordinates
 * Maps the viewing window onto the screen window (vp_matrix).
 ***********************************************************/
void            eval_viewport(vp_minx, vp_maxx, vp_miny, vp_maxy, vp_minz, vp_maxz)
double          vp_minx, vp_maxx;
double          vp_miny, vp_maxy;
double          vp_minz, vp_maxz;
{
   matrix          M;

   /****************************************
    * First, convert canonical view volume
    *  from perspective to parallel.
    ****************************************/
   zero_matrix(vp_matrix);
   vp_matrix[0][0] = 1.0;
   vp_matrix[1][1] = 1.0;
   vp_matrix[2][2] = 1.0 / (1.0 + Zmin);
   vp_matrix[2][3] = (0.0 - Zmin) / (1.0 + Zmin);
   vp_matrix[3][2] = -1.0;

   id_matrix(M);
   M[3][0] = 1.0;
   M[3][1] = 1.0;
   M[3][2] = 1.0;
   mult_matrix(vp_matrix, M, vp_matrix);
   id_matrix(M);
   M[0][0] = (vp_maxx - vp_minx) / 2;
   M[1][1] = (vp_maxy - vp_miny) / 2;
   M[2][2] = (vp_maxz - vp_minz);
   mult_matrix(vp_matrix, M, vp_matrix);
   id_matrix(M);
   M[3][0] = vp_minx;
   M[3][1] = vp_miny;
   M[3][2] = vp_minz;
   mult_matrix(vp_matrix, M, vp_matrix);

#ifdef TRANSFORM
   copy_matrix(M, vp_matrix);
   transpose(vp_matrix, M);
#endif

#ifdef DEBUG_VIEW
   printf("window to viewport matrix:\n");
   print_matrix(vp_matrix);
#endif
}

/*****************************************************
 * short pipe_vector( v1,v2)
 *  v1,v2 - line using the 'vector' data type
 * Puts the line through the rendering pipeline, clips
 * and returns TRUE to accept or FALSE if line is not
 * to be drawn.
 ****************************************************/
short           pipe_vector(v1, v2)
vector          v1, v2;
{

#ifdef DEBUG_PIPE
   printf("------------------- begin pipe:\n");
   print_vector(v1);
#endif
#ifndef TRANSFORM
   vector_transform(v1, v1, T_matrix);

#ifdef DEBUG_PIPE
   printf("after projection:\n");
   print_vector(v1);
#endif

   vector_transform(v2, v2, T_matrix);
#endif

   if (vector_clip(v1, v2))
   {
      vector_transform(v1, v1, vp_matrix);
      homogenize(v1);

      vector_transform(v2, v2, vp_matrix);
      homogenize(v2);

#ifdef DEBUG_PIPE
      printf("after clipping:\n");
      print_vector(v1);
#endif

      return (TRUE);
   }

#ifdef DEBUG_PIPE
   printf("not drawn!\n");
#endif

   return (FALSE);
}

/*****************************************************
 * short pipe_point( v)
 *  v - point using the 'vector' data type
 * Puts the point through the rendering pipeline,
 * clips and returns TRUE to accept or FALSE if
 * point is not to be drawn.
 ****************************************************/
short           pipe_point(v)
vector          v;
{
   short           rtn;

   vector_transform(v, v, T_matrix);
   homogenize(v);

#ifdef DEBUG_PIPE
   printf("after homogenization:\n");
   print_vector(v);
#endif

   if (rtn = point_clip(v))
   {
      vector_transform(v, v, vp_matrix);

#ifdef DEBUG_PIPE
      printf("drew dis sucka!\n");
      print_vector(v);
   }
   else
   {
      printf("did not draw dis sucka!\n");
#endif
   }
   return (rtn);
}

short           pipe_line(p, q)
point          *p, *q;
{
   vector          v1, v2;
   short           rtn;

   point2vector(v1, p);
   point2vector(v2, q);
   rtn = pipe_vector(v1, v2);
   vector2point(p, v1);
   vector2point(q, v2);

   return (rtn);
}

short           point_clip(p)
point          *p;
{
   if ((p->x) < Zdepth || (p->x) > -Zdepth)
      return (FALSE);
   else if ((p->y) < Zdepth || (p->y) > -Zdepth)
      return (FALSE);
   else if ((p->z) < -1.0 || (p->z) > Zmin)
      return (FALSE);
   return (TRUE);
}

/***************************************************
 * clip_3d - 3D implementation of Liang-Barsky
 *  parametric line-clipping algorithm.
 *  Code is derived from Foley, VanDam, etc.
 *  _Computer Graphics: Principles and Practice_,
 *  section 6.5.3.
 ***************************************************/
short           clip_t(denom, num, tE, tL)
double          denom, num;
double         *tE, *tL;
{
   double          t;
   short           accept = TRUE;

   if (denom > 0.0)
   {
      t = num / denom;
      if (t > *tL)
         accept = FALSE;
      else if (t > *tE)
         *tE = t;
   }
   else if (denom < 0.0)
   {
      t = num / denom;
      if (t < *tE)
         accept = FALSE;
      else if (t < *tL)
         *tL = t;
   }
   else if (num > 0.0)
      accept = FALSE;
   return (accept);
}

short           clip_3d(x0, y0, z0, x1, y1, z1, zmin)
double         *x0, *y0, *z0;
double         *x1, *y1, *z1;
double          zmin;
{
   short           accept = FALSE;
   double          dx, dy, dz;
   double          tmin = 0.0, tmax = 1.0;

   dx = (*x1) - (*x0);
   dz = (*z1) - (*z0);
   if (clip_t(-dx - dz, (*x0) + (*z0), &tmin, &tmax))
   {
      if (clip_t(dx - dz, -(*x0) + (*z0), &tmin, &tmax))
      {
         dy = (*y1) - (*y0);
         if (clip_t(dy - dz, -(*y0) + (*z0), &tmin, &tmax))
         {
            if (clip_t(-dy - dz, (*y0) + (*z0), &tmin, &tmax))
            {
               if (clip_t(-dz, (*z0) - zmin, &tmin, &tmax))
               {
                  if (clip_t(dz, -(*z0) - 1.0, &tmin, &tmax))
                  {
                     accept = TRUE;
                     if (tmax < 1.0)
                     {
                        *x1 = (*x0) + tmax * dx;
                        *y1 = (*y0) + tmax * dy;
                        *z1 = (*z0) + tmax * dz;
                     }
                     if (tmin > 0.0)
                     {
                        *x0 = (*x0) + tmin * dx;
                        *y0 = (*y0) + tmin * dy;
                        *z0 = (*z0) + tmin * dz;
                     }
                  }

#ifdef DEBUG_CLIP
                  else
                  {
                     printf("Rej by back: ");
                     printf("<%f %f %f> - <%f %f %f>\n",
                        *x0, *y0, *z0, *x1, *y1, *z1);
                  }
#endif
               }

#ifdef DEBUG_CLIP
               else
               {
                  printf("Rej by front: ");
                  printf("<%f %f %f> - <%f %f %f>\n",
                     *x0, *y0, *z0, *x1, *y1, *z1);
               }
#endif
            }

#ifdef DEBUG_CLIP
            else
            {
               printf("Rej by top: ");
               printf("<%f %f %f> - <%f %f %f>\n",
                  *x0, *y0, *z0, *x1, *y1, *z1);
            }
#endif
         }

#ifdef DEBUG_CLIP
         else
         {
            printf("Rej by bottom: ");
            printf("<%f %f %f> - <%f %f %f>\n",
               *x0, *y0, *z0, *x1, *y1, *z1);
         }
#endif
      }

#ifdef DEBUG_CLIP
      else
      {
         printf("Rej by left: ");
         printf("<%f %f %f> - <%f %f %f>\n",
            *x0, *y0, *z0, *x1, *y1, *z1);
      }
#endif
   }

#ifdef DEBUG_CLIP
   else
   {
      printf("Rej by right: ");
      printf("<%f %f %f> - <%f %f %f>\n",
         *x0, *y0, *z0, *x1, *y1, *z1);
   }
   if (accept)
   {
      printf("Accepted: ");
      printf("<%f %f %f> - <%f %f %f>\n",
         *x0, *y0, *z0, *x1, *y1, *z1);
   }
#endif

   return (accept);
}

short           vector_clip(v0, v1)
vector          v0, v1;
{
   double          x0, y0, z0;
   double          x1, y1, z1;

   vector2double(v0, &x0, &y0, &z0);
   vector2double(v1, &x1, &y1, &z1);
   if (clip_3d(&x0, &y0, &z0, &x1, &y1, &z1, Zmin))
   {
      double2vector(v0, x0, y0, z0);
      double2vector(v1, x1, y1, z1);
      return (TRUE);
   }
   else
      return (FALSE);
}
